/*
  Last changed Time-stamp: <2009-02-06 19:08:41 xtof>
  $Id: Interfaces.cpp,v 1.3 2009/02/06 18:17:45 xtof Exp $
*/
#include <cstring>
#include <string>
#include <vector>

// headers from nnet_1_5
#include "distances.h"
#include "neighbor_net.h"
#include "simple_nexus.h"

static std::vector<sequence>
nnet_encode_sequences(std::vector<std::string> seqs, data_types data_type,
		      std::string alphabet);

/**
 * Interface to nnet_1_5 from David Bryant <bryant@math.mcgill.ca>
 * returns the cyclic order of the sequences
 */
std::vector<int>
nnet_ordering(std::vector<std::string> seqs,
	      std::vector<std::vector<double> >& D,
	      char seqtype,
	      char* distmethod) {

  std::string alphabet;
  // compute distances
  std::vector<int> bs_sample; bs_sample.clear();

  // dispach type of sequence data
  data_types seq_type;
  switch(seqtype) {
  case 'D':
    seq_type = DNA;
    alphabet = DNA_STATES;
    break;
  case 'P':
    seq_type = Protein;
    alphabet = PROTEIN_STATES;
    break;
  case 'R':
    seq_type = RNA;
    alphabet = RNA_STATES;
    break;
  default:
    seq_type = standard;
  }

  alphabet += '-';

  // dispach distance method
  transform_types dist_method;
  if (strcmp(distmethod, "GTR")==0)
    dist_method = GTR;
  else
    dist_method = HAMMING;
  
  // convert sequences to format used by nnet_1_5  
  std::vector<sequence> enc_seqs = nnet_encode_sequences(seqs, seq_type,
							 alphabet);

  // compute distance matrix
  if (D.empty())
    compute_distance(enc_seqs, seq_type, D, dist_method, true,
		     0.0, 0.0, 0.0, 1.0, bs_sample);

  // compute network
  std::vector<int> ordering = std::vector<int>();
  neighbor_net(D, ordering);

  return (ordering);
}

/**
 * convert sequences to a nnet compliant form
 */
static std::vector<sequence>
nnet_encode_sequences(std::vector<std::string> seqs, data_types data_type,
		      std::string alphabet)
{
  std::vector<sequence> enc_seqs = std::vector<sequence>();
  
  for (std::vector<std::string>::iterator s = seqs.begin();
       s != seqs.end();
       s++) {
    
    sequence enc_s = sequence();
    for (std::string::iterator ch = s->begin();
	 ch != s->end();
	 ch++) {
      // encode character
      if (alphabet.find_first_of(*ch) == std::string::npos) {
	// std::cerr << "found " << *ch << std::endl;
	enc_s.push_back(get_state_id(data_type, MISSING));
      }
      else	
	enc_s.push_back(get_state_id(data_type, *ch));
    }

    enc_seqs.push_back(enc_s); 
  }

  return (enc_seqs);
}

// End of file
