/*
 *  distances.cpp
 *  nnet
 *
 *  Copyright (c) 2001 __MyCompanyName__. All rights reserved.
 *
 */

#include <cmath>

#include"distances.h"
    
//#define DEBUG_MATRIX
static double pairwise_distance(sequence& seq1,
                                sequence& seq2,
                                data_types seq_type,
                                transform_types transform_type,
                                bool equal_rates,
                                double shape,
                                double pinvar,
                                double switch_rate,
                                const std::vector<int>& bs_sample) {


    Array2D F,X;
    Array1D Pi,PiSqrt,d,e;
    int k,num_sites,num_not_missing;
    short int i,j,n;
    double D_ij,F_ij,Pi_i,x;
    bool bootstrapping = !(bs_sample.empty());
    
    if (seq_type == standard)
        n =2;
    else if ((seq_type == DNA) || (seq_type == RNA))
        n=4;
    else
        n=21;

    resize(F,n);
    num_not_missing = 0;

    num_sites = seq1.size();
    for(k = 0;k<num_sites; k++) {
        i = seq1[k];
        j = seq2[k];
        if (!state_missing(i) && !state_missing(j)) {
            if (!bootstrapping) {
                F[i][j]+=1.0;
                num_not_missing++;
            }
            else {
                F[i][j] += bs_sample[k];
                num_not_missing+=bs_sample[k];
            }
        }
    }

    if (num_not_missing == 0)
        return false;

#ifdef DEBUG_MATRIX
    cout<<"F matrix : \n";
    cout<<"F:=matrix([\n";
    for(i = 0 ; i< n; i++) {
        cout<<"[";
        for(j = 0 ; j<n; j++) {
            if (j>0) cout<<",";
            cout<<F[i][j];
        }
        cout<<"]";
        if (i<n-1) cout<<",";
        cout<<std::endl;
    }
#endif
    
  

    D_ij = -1;
    
    if (transform_type == HAMMING) {
        /* Return 1.0 - trace(F) */
        D_ij = num_not_missing;
        for(i=0;i<n;i++)
            D_ij -= F[i][i];
        D_ij = D_ij / num_not_missing;
        
    }
    else if (transform_type == DICE) {
        D_ij = 1.0 - 2.0*F[1][1] / (2.0 * F[1][1] + F[1][0] + F[0][1]);
    }
    else if (transform_type == JAKARD) {
        D_ij = 1.0 - F[1][1] / (F[1][1] + F[1][0] + F[0][1]);
    }
    else if ((transform_type == GTR)||(transform_type == COVGTR)) {

        /* Compute proportions and symmetrize matrix */
        for(i=0;i<n;i++) {
            for(j=0;j<=i;j++) {
                F_ij = 0.5*(F[i][j] + F[j][i]) /  (double) num_not_missing;
                F[i][j] = F[j][i] = F_ij;
            }
        }
        
        /* Compute base frequencies */
        resize(Pi,n);
        resize(PiSqrt,n);
        for(i=0;i<n;i++) {
            Pi_i = 0.0;
            for(j=0;j<n;j++)
                Pi_i += F[i][j];
            Pi[i]=Pi_i;
            PiSqrt[i] = sqrt(Pi_i);
        }

        /* F <-  Pi^{-1/2} F Pi^{-1/2} */
        for(i=0;i<n;i++) {
            F[i][i] = F[i][i]/Pi[i];
            for(j=0;j<i;j++) {
                F_ij = F[i][j]/PiSqrt[i]/PiSqrt[j];
                F[i][j] = F[j][i] = F_ij;
            }
        }
        
        /* Diagonalise F */
        Diagonalise(F,d,e);
        
        if (transform_type == GTR) {
            if (equal_rates || (shape==0.0)) {
                /* d = log(d) */
                for(i=0;i<n;i++) {
                    if (d[i] < 0.0)
                        return -1.0; /* Saturated */
                    else
                        d[i] = log(d[i]);
                }
            }
            else {
                for(i=0;i<n;i++) {
                    if (d[i] < 0.0)
                        return -1.0; /* Saturated */
                    else
                        d[i] = shape*(1.0-pow(d[i],-1.0/shape));
                }
            }
        }
#ifdef COVGTR
        else { /* COVGTR */
            double s1,s2;
            s1 =  switch_rate / (2.0 *(1.0 - pinvar));
            s2 =  switch_rate / (2.0 * pinvar);

            for(i=0;i<n;i++) {
                if (d[i] < 0.0)
                    return -1.0; /* Saturated */
                else
                    /* Find z such that M(z) = d[i] */
                    d[i] = inverseM(s1,s2,d[i]);
            }
        }
#endif
        /* Dist = - \sum_i Pi[i] \sum_j F_ij d_j F_ij */
        D_ij = 0.0;
        for(i=0;i<n;i++) {
            x = 0.0;
            for(j=0;j<n;j++) {
                F_ij = F[i][j];
                x += F_ij * F_ij * d[j];
            }
            D_ij -= Pi[i]* x;
        }
#ifdef DEBUG_MATRIX
        cout<<"Distance = "<<D_ij<<std::endl;
#endif
    }
    
    return D_ij;
}

static double EvalL(double x,
                    const sequence& seq1,
                    const sequence& seq2,
                    const std::vector<std::vector<double> >& site_freqs,
                    const std::vector<double>& freq_square)
{
    double y_x;
    int k;
    short int s1,s2;
    int nchar = seq1.size();
    y_x = 0.0;
    double f_vk,f_k;

    for(k=0;k<nchar;k++) {
        s1 = seq1[k];
        s2 = seq2[k];
        if (state_missing(s1)||state_missing(s2))
            continue;
        f_vk = site_freqs[k][s2];
        f_k = freq_square[k];
        if (s1 == s2) 
            y_x += log(f_vk + (1.0 - f_vk)*exp(-x/f_k));
        else
            y_x += log(f_vk*(1.0 - exp(-x/f_k)));
    }

    return -y_x;
}


/* Find a local minimum using (for now) the Golden Section method */
const double GS_EPSILON = 0.000001;
const double GS_MAXDIST = 10.0;

static double local_min_1D(double xmin,
                           double xmax,
                           const sequence& seq1,
                           const sequence& seq2,
                           const std::vector<std::vector<double> >& site_freqs,
                           const std::vector<double>& freq_square) {

    double a,b,tau,aa,bb,faa,fbb;
    tau = 2.0/(1.0 + sqrt(5.0));
    
    a = xmin;
    b = xmax;
    aa = a + (1.0 - tau)*(b-a);
    bb = a + tau*(b-a);
    faa = EvalL(aa,seq1,seq2,site_freqs,freq_square);
    fbb = EvalL(bb,seq1,seq2,site_freqs,freq_square);
    
    
    while ((b-a) > GS_EPSILON) {
       // cout<<"["<<a<<","<<aa<<","<<bb<<","<<b<<"] \t \t ("<<faa<<","<<fbb<<")"<<std::endl;

        if (faa < fbb) {
            b = bb;
            bb = aa;
            fbb = faa;
            aa = a+(1.0 - tau)*(b-a);
            faa = EvalL(aa,seq1,seq2,site_freqs,freq_square);
        }
        else {
            a = aa;
            aa = bb;
            faa = fbb;
            bb = a + tau*(b-a);
            fbb = EvalL(bb,seq1,seq2,site_freqs,freq_square);
        }
    }
    return a;
}

static bool ComputeSiteSpecific(const std::vector<sequence>& seqs,
				dbl_array& D, double ss_weight) {


    
  std::vector<std::vector<double> > site_freqs;
  std::vector<int> ungapped;
  std::vector<double> freq_square;
  std::vector<double> acid_freqs;
  int num_bases = 0;
    
    int i,j,k;
    int ntax,nchar;
    short int s_ij,s,s1,s2;
    double sf;

    ntax = seqs.size();
    nchar = seqs[0].size();

    /* For every site and every state we store the frequency of that state at that site
        and the number of non-gapped sequences at that site*/
    site_freqs.resize(nchar);
    ungapped.resize(nchar);
    freq_square.resize(nchar);
    acid_freqs.resize(20);
    fill(acid_freqs.begin(),acid_freqs.end(),0.0);

    /* First we form a table of all amino acid frequencies */
    for(i=0;i<ntax;i++) {
        for(j=0;j<nchar;j++) {
            s_ij = seqs[i][j];
            if (!state_missing(s_ij)) {
                acid_freqs[s_ij]+=1.0;
                num_bases++;
            }
        }
    }
    std::cout<<"ss_weight equals "<<ss_weight<<std::endl;
    
    std::cout<<"Frequencies are\n";

    for(s=0;s<20;s++) {
        acid_freqs[s] = acid_freqs[s]/(double)num_bases;
        std::cout<<get_state(Protein,s)<<":\t"<<acid_freqs[s]<<std::endl;
    }


    /* Now we compute the site specific frequencies */
    for(j=0;j<nchar;j++) {
        site_freqs[j].resize(20);
        ungapped[j] = 0;
        fill(site_freqs[j].begin(),site_freqs[j].end(),0);
        for(i=0;i<ntax;i++) {
            s_ij = seqs[i][j];
            if (!state_missing(s_ij)) {
                ungapped[j]++;
                site_freqs[j][s_ij]+=1.0;
            }
        }
        for(s=0;s<20;s++) {
            site_freqs[j][s] = site_freqs[j][s]/(double)ungapped[j];
          //  std::cout<<site_freqs[j][s]<<"\t";
        }
       // std::cout<<std::endl;
    }

    /* Now we replace the site frequencies with a convex combination of site frequencies and overall frequencies */
    for(j=0;j<nchar;j++) {
        for(s=0;s<20;s++) {
            site_freqs[j][s] = ss_weight * site_freqs[j][s] + (1.0 - ss_weight)*acid_freqs[s];
         //   std::cout<<site_freqs[j][s]<<"\t";
        }
       // std::cout<<std::endl;
    }

    /* Now we compute 1.0 minus the sum of the squares of the frequencies for each site */

    for(j=0;j<nchar;j++) {
        freq_square[j] = 1.0;
        for(s=0;s<20;s++) {
            sf = site_freqs[j][s];
            freq_square[j] -= sf*sf;
        }
    }
 
    
    /* We now fill in the distance matrix */
    for(i=0;i<ntax;i++) {
        for(j=0;j<i;j++) {
            /* We need to check to see whether the two sequences have distance > 0 to avoid taking the log of zero*/
            bool has_diff = false;
            for (k=0;k<nchar;k++) {
                s1 = seqs[i][k];
                s2 = seqs[j][k];
                if (!state_missing(s1)&&!state_missing(s2)&&(s1!=s2)) {
                    has_diff = true;
                    break;
                }
            }

            if (!has_diff)
                D[i][j] = D[j][i] = 0.0;
            else 
                D[i][j] = local_min_1D(GS_EPSILON,GS_MAXDIST,seqs[i],seqs[j],site_freqs,freq_square);
            std::cout<<D[i][j]<<"\t";
        }
        std::cout<<std::endl;
    }
    

    return true;
}

//TO DO: CHANGE PARAMETERS TO A RECORD 
bool compute_distance(std::vector<sequence> & seqs,
                      data_types seq_type,
                      dbl_array& D,
                      transform_types transform_type,
                      bool equal_rates,
                      double shape,
                      double pinvar,
                      double switch_rate,
                      double ss_weight,
                      const std::vector<int>& bs_sample) {

    int ntax = seqs.size();
    int i,j;

    D.resize(ntax);
    for(i=0;i<ntax;i++)
        D[i].resize(ntax);


    /* DICE and JAKARD only defined for binary sequences */
    if ((transform_type == DICE) || (transform_type == JAKARD))
        if (seq_type!=standard)
            return false;
    
    /* SITESPEC distance defined only for amino acide distances */
    if (transform_type == SITESPEC) {
        if (seq_type!=Protein) 
            return false;
        return ComputeSiteSpecific(seqs,D,ss_weight);
    }
    
    
    
    
    for(i=0;i<ntax;i++) {
        for(j=0;j<i;j++) {
        //    std::cout<<"Computing D["<<i<<","<<j<<"]"<<std::endl;
            D[i][j] = D[j][i] = pairwise_distance(seqs[i],seqs[j],seq_type,transform_type,equal_rates,shape,pinvar,switch_rate, bs_sample);
          //  std::cout<<" D["<<i<<","<<j<<"] = "<<D[i][j]<<std::endl;

        }
    }
#ifdef DEBUG_MATRIX
    std::cout<<"\n\n\n Distances: \n";
    for(i=0;i<ntax;i++) {
        for(j=0;j<i;j++)
            std::cout<<D[i][j]<<"\t";
	std::cout<<std::endl;
    }
#endif
    
    return true;
}




    
