/*
  Last changed Time-stamp: <2009-02-06 18:40:36 xtof>
  $Id: holders.cpp,v 1.2 2009/02/06 17:41:20 xtof Exp $
*/
#include <algorithm>
#include <iostream>

#include "holders.h"
#include "CyclicOrderer.h"
#include "Utils.h"

using namespace std;

// *** WHolder *** //

WHolder::WHolder (INTINTVEC theLists, int N, QUARTET theQuartetWeights)
{
  counts.resize (over4 (N));
  weights.resize (over4 (N));

  for (int n=0; n<over4 (N); n++) {
    counts[n].A=0;
    counts[n].B=0;
    counts[n].C=0;
    weights[n].A=0.0;
    weights[n].B=0.0;
    weights[n].C=0.0;
  }

  // now, we store everything properly
  for (int i=1; i<N-2; i++) {
    for (int j=i+1; j<N-1; j++) {
      for (int k=j+1; k<N; k++) {
	for (int l=k+1; l<N+1; l++) {
	  // for every pair i, j
	  // find which list they are in
	  // for those two indices
	  // take all combinations of path entries
	  // against all combinations of other paths by all other path entries
	  int a=-1, b=-1, c=-1, d=-1;

	  for (unsigned int m=0; m<theLists.size (); m++) {
	    INTVEC tL=theLists[m];
	    if (count (tL.begin(), tL.end(), i)) {
	      a=m;
	      break;
	    }
	  }
	  
	  for (unsigned int m=0; m<theLists.size (); m++) {
	    INTVEC tL=theLists[m];
	    if (count (tL.begin(), tL.end(), j)) {
	      b=m;
	      break;
	    }
	  }
	  
	  for (unsigned int m=0; m<theLists.size (); m++) {
	    INTVEC tL=theLists[m];
	    if (count (tL.begin(), tL.end(), k)) {
	      c=m;
	      break;
	    }
	  }
	  
	  for (unsigned int m=0; m<theLists.size (); m++) {
	    INTVEC tL=theLists[m];
	    if (count (tL.begin(), tL.end(), l)) {
	      d=m;
	      break;
	    }
	  }
	  
	  if (a==b || a==c || a==d || b==c || b==d || c==d) {
	    // if on the same path, no quartets meet the conditions
	    int idx=over4 (l-1)+over3 (k-1)+over2 (j-1)+over1 (i-1);

	    counts[idx].A=0;
	    counts[idx].B=0;
	    counts[idx].C=0;
	    weights[idx].A=0.0;
	    weights[idx].B=0.0;
	    weights[idx].C=0.0;
	    
	    continue;
	  }

	  // otherwise:
	  // we now have the list indices
	  int count1=0;
	  double weight1=0.0;
	  int count2=0;
	  double weight2=0.0;
	  int count3=0;
	  double weight3=0.0;

	  INTVEC A=theLists[a];
	  INTVEC B=theLists[b];
	  INTVEC C=theLists[c];
	  INTVEC D=theLists[d];

	  // we now have four non-same lists
	  for (unsigned int xA=0; xA<A.size(); xA++) {
	    for (unsigned int xB=0; xB<B.size(); xB++) {
	      for (unsigned int xC=0; xC<C.size(); xC++) {
		for (unsigned int xD=0; xD<D.size(); xD++) {
		  
		  // this is a unique, suitable quartet
		  int yA=A[xA];
		  int yB=B[xB];
		  int yC=C[xC];
		  int yD=D[xD];

		  count1++;
		  weight1+=getQuartetWeights (theQuartetWeights, yA, yB, yC, yD);
		  count2++;
		  weight2+=getQuartetWeights (theQuartetWeights, yA, yC, yB, yD);
		  count3++;
		  weight3+=getQuartetWeights (theQuartetWeights, yA, yD, yB, yC);
		}
	      }
	    }
	  }
	  
	  int idx=over4 (l-1)+over3 (k-1)+over2 (j-1)+over1 (i-1);
	  counts[idx].A=count1;
	  counts[idx].B=count2;
	  counts[idx].C=count3;
	  weights[idx].A=weight1;
	  weights[idx].B=weight2;
	  weights[idx].C=weight3;
	}
      }
    }
  }
}

int WHolder::getN (int i, int j, int k, int l)
{
  /**
     Create size-ordered quadruple x, y, u, v from i, j, k, l
  */

  int position=0;

  if ((i>k && j>k && i>l && j>l) ||
      (i<k && j<k && i<l && j<l)) {
    // both largest to the left or to the right
    position=1;
  }
  else if (((i>k && i>l && ((j>l && j<k) || (j<l && j>k))) ||
	    (j>k && j>l && ((i>l && i<k) || (i<l && i>k)))) ||
	   (l>j && l>i && ((k>i && k<j) || (k<i && k>j))) ||
	   (k>j && k>i && ((l>i && l<j) || (l<i && l>j)))) {
    // largest and third to the left or to the right
    position=2;
  }
  else if ((i>k && i>l && (j<l && j<k)) ||
	   (j>k && j>l && (i<l && i<k)) ||
	   (l>j && l>i && (k<i && k<j)) ||
	   (k>j && k>i && (l<i && l<j))) {
    // largest and smallest to the left or to the right
    position=3;
  }

  int x=i;
  int y=j;
  int u=k;
  int v=l;
  int m;

  if (y>x) {
    m=x;
    x=y;
    y=m;
  }
  if (u>x) {
    m=x;
    x=u;
    u=m;
  }
  if (v>x) {
    m=x;
    x=v;
    v=m;
  }
  if (u>y) {
    m=y;
    y=u;
    u=m;
  }
  if (v>y) {
    m=y;
    y=v;
    v=m;
  }
  if (v>u) {
    m=u;
    u=v;
    v=m;
  }

  int idx=over4 (x-1)+over3 (y-1)+over2 (u-1)+over1 (v-1);
  
  if (position==1) {
    return (counts[idx].A);
  }
  else if (position==2) {
    return (counts[idx].B);
  }
  else if (position==3) {
    return (counts[idx].C);
  }
  else return 0;
}

void WHolder::setN (int i, int j, int k, int l, int newN)
{
  /**
     Create size-ordered quadruple x, y, u, v from i, j, k, l
  */

  int position=0;

  if ((i>k && j>k && i>l && j>l) ||
      (i<k && j<k && i<l && j<l)) {
    // both largest to the left or to the right
    position=1;
  }
  else if (((i>k && i>l && ((j>l && j<k) || (j<l && j>k))) ||
	    (j>k && j>l && ((i>l && i<k) || (i<l && i>k)))) ||
	   (l>j && l>i && ((k>i && k<j) || (k<i && k>j))) ||
	   (k>j && k>i && ((l>i && l<j) || (l<i && l>j)))) {
    // largest and third to the left or to the right
    position=2;
  }
  else if ((i>k && i>l && (j<l && j<k)) ||
	   (j>k && j>l && (i<l && i<k)) ||
	   (l>j && l>i && (k<i && k<j)) ||
	   (k>j && k>i && (l<i && l<j))) {
    // largest and smallest to the left or to the right
    position=3;
  }

  int x=i;
  int y=j;
  int u=k;
  int v=l;
  int m;

  if (y>x) {
    m=x;
    x=y;
    y=m;
  }
  if (u>x) {
    m=x;
    x=u;
    u=m;
  }
  if (v>x) {
    m=x;
    x=v;
    v=m;
  }
  if (u>y) {
    m=y;
    y=u;
    u=m;
  }
  if (v>y) {
    m=y;
    y=v;
    v=m;
  }
  if (v>u) {
    m=u;
    u=v;
    v=m;
  }

  int idx=over4 (x-1)+over3 (y-1)+over2 (u-1)+over1 (v-1);
  
  if (position==1) {
    counts[idx].A=newN;
  }
  else if (position==2) {
    counts[idx].B=newN;
  }
  else if (position==3) {
    counts[idx].C=newN;
  }
  else {
    /* nothing */
  }
}

double WHolder::getW (int i, int j, int k, int l)
{
  /**
     Create size-ordered quadruple x, y, u, v from i, j, k, l
  */

  int position=0;

  if ((i>k && j>k && i>l && j>l) ||
      (i<k && j<k && i<l && j<l)) {
    // both largest to the left or to the right
    position=1;
  }
  else if (((i>k && i>l && ((j>l && j<k) || (j<l && j>k))) ||
	    (j>k && j>l && ((i>l && i<k) || (i<l && i>k)))) ||
	   (l>j && l>i && ((k>i && k<j) || (k<i && k>j))) ||
	   (k>j && k>i && ((l>i && l<j) || (l<i && l>j)))) {
    // largest and third to the left or to the right
    position=2;
  }
  else if ((i>k && i>l && (j<l && j<k)) ||
	   (j>k && j>l && (i<l && i<k)) ||
	   (l>j && l>i && (k<i && k<j)) ||
	   (k>j && k>i && (l<i && l<j))) {
    // largest and smallest to the left or to the right
    position=3;
  }

  int x=i;
  int y=j;
  int u=k;
  int v=l;
  int m;

  if (y>x) {
    m=x;
    x=y;
    y=m;
  }
  if (u>x) {
    m=x;
    x=u;
    u=m;
  }
  if (v>x) {
    m=x;
    x=v;
    v=m;
  }
  if (u>y) {
    m=y;
    y=u;
    u=m;
  }
  if (v>y) {
    m=y;
    y=v;
    v=m;
  }
  if (v>u) {
    m=u;
    u=v;
    v=m;
  }

  int idx=over4 (x-1)+over3 (y-1)+over2 (u-1)+over1 (v-1);
  
  if (position==1) {
    return (weights[idx].A);
  }
  else if (position==2) {
    return (weights[idx].B);
  }
  else if (position==3) {
    return (weights[idx].C);
  }
  else return 0;
}

void WHolder::setW (int i, int j, int k, int l, double newW)
{
  /**
     Create size-ordered quadruple x, y, u, v from i, j, k, l
  */

  int position=0;

  if ((i>k && j>k && i>l && j>l) ||
      (i<k && j<k && i<l && j<l)) {
    // both largest to the left or to the right
    position=1;
  }
  else if (((i>k && i>l && ((j>l && j<k) || (j<l && j>k))) ||
	    (j>k && j>l && ((i>l && i<k) || (i<l && i>k)))) ||
	   (l>j && l>i && ((k>i && k<j) || (k<i && k>j))) ||
	   (k>j && k>i && ((l>i && l<j) || (l<i && l>j)))) {
    // largest and third to the left or to the right
    position=2;
  }
  else if ((i>k && i>l && (j<l && j<k)) ||
	   (j>k && j>l && (i<l && i<k)) ||
	   (l>j && l>i && (k<i && k<j)) ||
	   (k>j && k>i && (l<i && l<j))) {
    // largest and smallest to the left or to the right
    position=3;
  }

  int x=i;
  int y=j;
  int u=k;
  int v=l;
  int m;

  if (y>x) {
    m=x;
    x=y;
    y=m;
  }
  if (u>x) {
    m=x;
    x=u;
    u=m;
  }
  if (v>x) {
    m=x;
    x=v;
    v=m;
  }
  if (u>y) {
    m=y;
    y=u;
    u=m;
  }
  if (v>y) {
    m=y;
    y=v;
    v=m;
  }
  if (v>u) {
    m=u;
    u=v;
    v=m;
  }

  int idx=over4 (x-1)+over3 (y-1)+over2 (u-1)+over1 (v-1);
  
  if (position==1) {
    weights[idx].A=newW;
  }
  else if (position==2) {
    weights[idx].B=newW;
  }
  else if (position==3) {
    weights[idx].C=newW;
  }
  else {
    /* nothing */
  }
}
 
#if 0
WHolder::WHolder (INTINTVEC theLists, int N, QUARTET theQuartetWeights)
{
  counts.resize (over4 (N));
  weights.resize (over4 (N));

  for (int n=0; n<over4 (N); n++) {
    counts[n]=0;
    weights[n]=0.0;
  }

  // now, we store everything properly
  for (int i=1; i<N-2; i++) {
    for (int j=i+1; j<N-1; j++) {
      for (int k=j+1; k<N; k++) {
	for (int l=k+1; l<N+1; l++) {
	  // for every pair i, j
	  // find which list they are in
	  // for those two indices
	  // take all combinations of path entries
	  // against all combinations of other paths by all other path entries
	  
	  int a=-1, b=-1, c=-1, d=-1;
	  
	  for (unsigned int m=0; m<theLists.size (); m++) {
	    INTVEC tL=theLists[m];
	    
	    if (count (tL.begin(), tL.end(), i)) {
	      a=m;
	      break;
	    }
	  }
	  
	  for (unsigned int m=0; m<theLists.size (); m++) {
	    INTVEC tL=theLists[m];
	    
	    if (count (tL.begin(), tL.end(), j)) {
	      b=m;
	      break;
	    }
	  }
	  
	  for (unsigned int m=0; m<theLists.size (); m++) {
	    INTVEC tL=theLists[m];
	    
	    if (count (tL.begin(), tL.end(), k)) {
	      c=m;
	      break;
	    }
	  }
	  
	  for (unsigned int m=0; m<theLists.size (); m++) {
	    INTVEC tL=theLists[m];
	    
	    if (count (tL.begin(), tL.end(), l)) {
	      d=m;
	      break;
	    }
	  }

	  if (a==b || a==c || a==d || b==c || b==d || c==d) {
	    // if on the same path, no quartets meet the conditions
	    int idx=over4 (l-1)+over3 (k-1)+over2 (j-1)+over1 (i-1);
	    counts[idx]=0;
	    weights[idx]=0.0f;
	    
	    continue;
	  }

	  // otherwise:
	  // we now have the list indices
	  
	  int count=0; double weight=0.0;

	  INTVEC A=theLists[a];
	  INTVEC B=theLists[b];
	  INTVEC C=theLists[c];
	  INTVEC D=theLists[d];
	  
	  // we now have four non-same lists
	  for (unsigned int xA=0; xA<A.size(); xA++) {
	    for (unsigned int xB=0; xB<B.size(); xB++) {
	      for (unsigned int xC=0; xC<C.size(); xC++) {
		for (unsigned int xD=0; xD<D.size(); xD++) {

		  // this is a unique, suitable quartet
		  int yA=A[xA];
		  int yB=B[xB];
		  int yC=C[xC];
		  int yD=D[xD];

		  count++;
		  weight+=getQuartetWeights (theQuartetWeights, yA, yB, yC, yD);
		}
	      }
	    }
	  }

	  int idx=over4 (l-1)+over3 (k-1)+over2 (j-1)+over1 (i-1);
	  counts[idx]=count;
	  weights[idx]=weight;
	}
      }
    }
  }
}

int WHolder::getN (int i, int j, int k, int l)
{
  /**
     Create size-ordered quadruple x, y, u, v from i, j, k, l
  */
  int x=i;
  int y=j;
  int u=k;
  int v=l;
  int m;

  if (y>x) {
    m=x;
    x=y;
    y=m;
  }

  if (u>x) {
    m=x;
    x=u;
    u=m;
  }
  
  if (v>x) {
    m=x;
    x=v;
    v=m;
  }

  if (u>y) {
    m=y;
    y=u;
    u=m;
  }
  
  if (v>y) {
    m=y;
    y=v;
    v=m;
  }

  if (v>u) {
    m=u;
    u=v;
    v=m;
  }

  int idx=over4 (x-1)+over3 (y-1)+over2 (u-1)+over1 (v-1);

  return counts[idx];
}

void WHolder::setN (int i, int j, int k, int l, int newN)
{
  /**
     Create size-ordered quadruple x, y, u, v from i, j, k, l
  */
  int x=i;
  int y=j;
  int u=k;
  int v=l;
  int m;

  if (y>x) {
    m=x;
    x=y;
    y=m;
  }

  if (u>x) {
    m=x;
    x=u;
    u=m;
  }

  if (v>x) {
    m=x;
    x=v;
    v=m;
  }

  if (u>y) {
    m=y;
    y=u;
    u=m;
  }

  if (v>y) {
    m=y;
    y=v;
    v=m;
  }

  if (v>u) {
    m=u;
    u=v;
    v=m;
  }

  int idx=over4 (x-1)+over3 (y-1)+over2 (u-1)+over1 (v-1);

  counts[idx]=newN;
}

double WHolder::getW (int i, int j, int k, int l)
{
  /**
     Create size-ordered quadruple x, y, u, v from i, j, k, l
  */
  
  int x=i;
  int y=j;
  int u=k;
  int v=l;
  int m;

  if (y>x) {
    m=x;
    x=y;
    y=m;
  }

  if (u>x) {
    m=x;
    x=u;
    u=m;
  }

  if (v>x) {
    m=x;
    x=v;
    v=m;
  }

  if (u>y) {
    m=y;
    y=u;
    u=m;
  }

  if (v>y) {
    m=y;
    y=v;
    v=m;
  }

  if (v>u) {
    m=u;
    u=v;
    v=m;
  }

  int idx=over4 (x-1)+over3 (y-1)+over2 (u-1)+over1 (v-1);
  
  return weights [idx];
}

void WHolder::setW (int i, int j, int k, int l, double newW)
{
  /**
     Create size-ordered quadruple x, y, u, v from i, j, k, l
  */

  int x=i;
  int y=j;
  int u=k;
  int v=l;
  int m;

  if (y>x) {
    m=x;
    x=y;
    y=m;
  }

  if (u>x) {
    m=x;
    x=u;
    u=m;
  }

  if (v>x) {
    m=x;
    x=v;
    v=m;
  }

  if (u>y) {
    m=y;
    y=u;
    u=m;
  }

  if (v>y) {
    m=y;
    y=v;
    v=m;
  }

  if (v>u) {
    m=u;
    u=v;
    v=m;
  }

  int idx=over4 (x-1)+over3 (y-1)+over2 (u-1)+over1 (v-1);
  weights[idx]=newW;
}
#endif

// *** end of WHolder *** //

// *** NSHolder *** //

NSHolder::NSHolder (INTINTVEC theLists, int N, QUARTET theQuartetWeights)
{
  counts.resize (over2 (N));
  weights.resize (over2 (N));

  for (int n=0; n<over2 (N); n++) {
    counts[n]=0;
    weights[n]=0.0;
  }
  
  // now, we store everything properly
  for (int i=1; i<N; i++) {
    for (int j=i+1; j<N+1; j++) {
      // for every pair i, j
      // find which list they are in
      // for those two indices
      // take all combinations of path entries
      // against all combinations of other paths by all other path entries

      int a=-1, b=-1;

      for (unsigned int m=0; m<theLists.size (); m++) {
	INTVEC tL=theLists[m];
	
	if (count (tL.begin (), tL.end (), i)) {
	  a=m;
	  break;
	}
      }
  
      for (unsigned int m=0; m<theLists.size (); m++) {
	INTVEC tL=theLists[m];
	
	if (count (tL.begin (), tL.end (), j)) {
	  b=m;
	  break;
	}
      }

      if (a==b) {
	// if on the same path, no quartets meet the conditions
	int idx=over2 (j-1)+over1 (i-1);
	counts[idx]=0;
	weights[idx]=0.0;
	continue;
      }

      // otherwise:
      // we now have the list indices
      int count=0;
      double weight=0.0;

      INTVEC A=theLists[a];
      INTVEC B=theLists[b];

      for (int c=0; c<static_cast<int>(theLists.size ())-1; c++) {
	for (int d=c+1; d<static_cast<int>(theLists.size ()); d++) {
	  if (c!=a && c!=b && d!=a && d!=b) {
	    INTVEC C=theLists[c];
	    INTVEC D=theLists[d];
	    
	    // we now have four non-same lists
	    for (unsigned int xA=0; xA<A.size (); xA++) {
	      for (unsigned int xB=0; xB<B.size (); xB++) {
		for (unsigned int xC=0; xC<C.size (); xC++) {
		  for (unsigned int xD=0; xD<D.size (); xD++) {
		    
		    // this is a unique, suitable quartet
		    int yA=A[xA];
		    int yB=B[xB];
		    int yC=C[xC];
		    int yD=D[xD];
		    count++;
		    weight+=getQuartetWeights (theQuartetWeights, yA, yB, yC, yD);
		  }
		}
	      }
	    }
	  }
	}
      }
      int idx=over2 (j-1)+over1 (i-1);
      counts[idx]=count;
      weights[idx]=weight;
    }
  }
}

int NSHolder::getN (int i, int j)
{
  int x=max (i, j);
  int y=min (i, j);
  
  int idx=over2 (x-1)+over1 (y-1);
    
  return counts[idx];
}

void NSHolder::setN (int i, int j, int newN)
{
  int x=max (i, j);
  int y=min (i, j);
  
  int idx=over2 (x-1)+over1 (y-1);
  
  counts[idx]=newN;
}

double NSHolder::getS (int i, int j)
{
  int x=max (i, j);
  int y=min (i, j);
  
  int idx=over2 (x-1)+over1 (y-1);
    
  return weights[idx];
}

void NSHolder::setS (int i, int j, double newS)
{
  int x=max (i, j);
  int y=min (i, j);
  
  int idx=over2 (x-1)+over1 (y-1);
  
  weights[idx]=newS;
}

// *** end of NSHolder *** //


// *** THolder *** //

THolder::THolder (INTINTVEC theLists, int N, QUARTET theQuartetWeights)
{
  counts.resize (N);
  weights.resize (N);

  for (int n1=0; n1<N; n1++) {
    counts[n1].resize (N);
    weights[n1].resize (N);
    for (int n2=0; n2<N; n2++) {
      counts[n1][n2].resize (N);
      weights[n1][n2].resize (N);
      for (int n3=0; n3<N; n3++) {
	counts[n1][n2][n3]=0;
	weights[n1][n2][n3]=0.0;
      }
    }
  }
	
  for (int i=1; i<N+1; i++) {
    for (int j=1; j<N+1; j++) {
      // here, we must have that i and j are not on the same path
      // and we want their paths
      
      int a=-1, b=-1;

      for (unsigned int m=0; m<theLists.size(); m++) {
	INTVEC tL=theLists[m];
	
	if (count (tL.begin(), tL.end(), i)) {
	  a=m;
	  break;
	}
      }
      
      for (unsigned int m=0; m<theLists.size(); m++) {
	INTVEC tL=theLists[m];
	
	if (count (tL.begin(), tL.end(), j)) {
	  b=m;
	  break;
	}
      }

      if (a==b) {
	// if on the same path, no quartets meet the conditions
	continue;
      }

      INTVEC A=theLists[a];
      INTVEC B=theLists[b];

      for (int k=1; k<N+1; k++) {
	int c=-1;
	
	// and we must have that k is not on the path of i, j
	for (unsigned int m=0; m<theLists.size (); m++) {
	  INTVEC tL=theLists[m];
	  
	  if (count (tL.begin(), tL.end(), k)) {
	    c=m;
	    break;
	  }
	}

	if (c!=a && c!=b) {
	  // yay! uniqueness
	  int count=0;
	  double weight=0.0;
	  INTVEC C=theLists[c];
	  
	  for (unsigned int xA1=0; xA1<A.size()-1; xA1++) {
	    for (unsigned int xA2=xA1+1; xA2<A.size(); xA2++) {
	      for (unsigned int xB=0; xB<B.size(); xB++) {
		for (unsigned int xC=0; xC<C.size(); xC++) {
		  // this is a unique, suitable quartet
		  int yA1=A[xA1];
		  int yA2=A[xA2];
		  int yB=B[xB];
		  int yC=C[xC];

		  count++;
		  weight+=getQuartetWeights (theQuartetWeights, yA1, yB, yA2, yC);
		}
	      }
	    }
	  }
	  counts[i-1][j-1][k-1]=count;
	  weights[i-1][j-1][k-1]=weight;
	}
      }
    }
  }
}

int THolder::getN (int i, int j, int k)
{
  return counts[i-1][j-1][k-1];
}

void THolder::setN (int i, int j, int k, int newN)
{
  counts[i-1][j-1][k-1]=newN;
}

double THolder::getT (int i, int j, int k)
{
  return weights[i-1][j-1][k-1];
}

void THolder::setT (int i, int j, int k, double newT)
{
  weights[i-1][j-1][k-1]=newT;
}

// *** end of THolder ** //

// *** U0Holder *** //

U0Holder::U0Holder (INTINTVEC theLists, int N, QUARTET theQuartetWeights)
{
  counts.resize (over2 (N));
  weights.resize (over2 (N));
  
  for (int n=0; n<over2 (N); n++) {
    counts[n]=0;
    weights[n]=0.0;
  }
  
  // now, we store everything properly
  for (int i=1; i<N; i++) {
    for (int j=i+1; j<N+1; j++) {
      // for every pair i, j
      // find which list they are in
      // for those two indices
      // take all combinations of path entries
      // against all combinations of other paths by all other path entries

      int a=-1, b=-1;

      for (unsigned int m=0; m<theLists.size (); m++) {
	INTVEC tL=theLists[m];

	if (count (tL.begin(), tL.end(), i)) {
	  a=m;
	  break;
	}
      }
 
      for (unsigned int m=0; m<theLists.size (); m++) {
	INTVEC tL=theLists[m];

	if (count (tL.begin(), tL.end(), j)) {
	  b=m;
	  break;
	}
      }

      if (a==b) {
	// if on the same path, no quartets meet the conditions
	int idx=over2 (j-1)+over1 (i-1);
	counts[idx]=0;
	weights[idx]=0.0;
	continue;
      }

      // otherwise:
      // we now have the list indices
      int count=0;
      double weight=0.0;

      INTVEC A=theLists[a];
      INTVEC B=theLists[b];

      // we now have two non-same lists
      for (unsigned int xA1=0; xA1<A.size()-1; xA1++) {
	for (unsigned int xA2=xA1+1; xA2<A.size(); xA2++) {
	  for (unsigned int xB1=0; xB1<B.size ()-1; xB1++) {
	    for (unsigned int xB2=xB1+1; xB2<B.size(); xB2++) {
	      // this is a unique, suitable quartet
	      int yA1=A[xA1];
	      int yA2=A[xA2];
	      int yB1=B[xB1];
	      int yB2=B[xB2];
	      
	      count++;
	      weight+=getQuartetWeights (theQuartetWeights, yA1, yB1, yA2, yB2);
	      
	      // hope this does not mean doing stuff twice
	      count++;
	      weight+=getQuartetWeights (theQuartetWeights, yA2, yB2, yA1, yB1);
	    }
	  }
	}
      }

      int idx=over2 (j-1)+over1 (i-1);
      counts[idx]=count;
      weights[idx]=weight;
    }
  }
}

int U0Holder::getN (int i, int j)
{
  int x=max (i, j);
  int y=min (i, j);
  int idx=over2 (x-1)+over1 (y-1);

  return counts[idx];
}

void U0Holder::setN (int i, int j, int newN)
{
  int x=max (i, j);
  int y=min (i, j);
  int idx=over2 (x-1)+over1 (y-1);

  counts[idx]=newN;
}

double U0Holder::getU (int i, int j)
{
  int x=max (i, j);
  int y=min (i, j);
  int idx=over2 (x-1)+over1 (y-1);

  return weights[idx];
}

void U0Holder::setU (int i, int j, double newU)
{
  int x=max (i, j);
  int y=min (i, j);
  int idx=over2 (x-1)+over1 (y-1);

  weights[idx]=newU;
}

// *** end of U0Holder ** //

// *** U1Holder ** //

U1Holder::U1Holder (INTINTVEC theLists, int N, QUARTET theQuartetWeights)
{
  counts.resize (over2 (N));
  weights.resize (over2 (N));

  for (int n=0; n<over2 (N); n++) {
    counts[n]=0;
    weights[n]=0.0;
  }

  // now, we store everything properly
  for (int i=1; i<N; i++) {
    for (int j=i+1; j<N+1; j++) {
      // for every pair i, j
      // find which list they are in
      // for those two indices
      // take all combinations of path entries
      // against all combinations of other paths by all other path entries
      int a=-1, b=-1;

      for (unsigned int m=0; m<theLists.size (); m++) {
	INTVEC tL=theLists[m];
	if (count (tL.begin(), tL.end(), i)) {
	  a=m;
	  break;
	}
      }
 
      for (unsigned int m=0; m<theLists.size (); m++) {
	INTVEC tL=theLists[m];
	if (count (tL.begin(), tL.end(), j)) {
	  b=m;
	  break;
	}
      }
      
      if (a==b) {
	// if on the same path, no quartets meet the conditions
	int idx=over2 (j-1)+over1 (i-1);
	counts[idx]=0;
	weights[idx]=0.0;
	continue;
      }

      // otherwise:
      // we now have the list indices
      int count=0;
      double weight=0.0;

      INTVEC A=theLists[a];
      INTVEC B=theLists[b];

      // we now have two non-same lists
      for (unsigned int xA1=0; xA1<A.size()-1; xA1++) {
	for (unsigned int xA2=xA1+1; xA2<A.size(); xA2++) {
	  for (unsigned int xB1=0; xB1<B.size()-1; xB1++) {
	    for (unsigned int xB2=xB1+1; xB2<B.size(); xB2++) {
	      // this is a unique, suitable quartet
	      int yA1=A[xA1];
	      int yA2=A[xA2];
	      int yB1=B[xB1];
	      int yB2=B[xB2];
	      
	      count++;
	      weight+=getQuartetWeights (theQuartetWeights, yA1, yB2, yA2, yB1);
	      
	      // hope this does not mean doing stuff twice
	      count++;
	      weight+=getQuartetWeights (theQuartetWeights, yA2, yB1, yA1, yB2);
	    }
	  }
	}
      }
      int idx=over2 (j-1)+over1 (i-1);
      counts[idx]=count;
      weights[idx]=weight;
    }
  }
}

int U1Holder::getN (int i, int j)
{
  int x=max (i, j);
  int y=min (i, j);
  int idx=over2 (x-1)+over1 (y-1);
  
  return counts[idx];
}

void U1Holder::setN (int i, int j, int newN)
{
  int x=max (i, j);
  int y=min (i, j);
  int idx=over2 (x-1)+over1 (y-1);
  
  counts[idx]=newN;
}

double U1Holder::getU (int i, int j)
{
  int x=max (i, j);
  int y=min (i, j);
  int idx=over2 (x-1)+over1 (y-1);
  
  return weights[idx];
}

void U1Holder::setU (int i, int j, double newU)
{
  int x=max (i, j);
  int y=min (i, j);
  int idx=over2 (x-1)+over1 (y-1);
  
  weights[idx]=newU;
}

// *** end of U1Holder *** //

// *** ZHolder *** //

ZHolder::ZHolder (INTINTVEC theLists, int N)
{
  sizes.clear ();
  
  for (int i=1; i<N+1; i++) {
    sizes.push_back (z (theLists, i));
  }
}

int ZHolder::getZ (int i)
{
  return (sizes[i-1]);
}

void ZHolder::setZ (int i, int newZ)
{
  sizes[i-1]=newZ;
}

int ZHolder::z (INTINTVEC theLists, int i)
{
  for (unsigned int n=0; n<theLists.size (); n++) 
    {
      INTVEC tL=theLists[n];
    
      if (count (tL.begin(), tL.end(), i)) 
	{
	  if (i==tL[0]) return tL.size ();
	}
    }
  return 0;
}

// *** end of ZHolder *** //
