/*
  Last changed Time-stamp: <2010-02-18 14:09:00 xtof>
  $Id: options.c,v 1.8 2010/02/18 15:07:03 xtof Exp $
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <getopt.h>
#include "options.h"

static OptionS OptS;

static struct option const long_options[] =
{ {"Kaffee" ,    no_argument,       0,   0},
  {"cutoff" ,    required_argument, 0,   0},
  {"matrix",     required_argument, 0,   0},
  {"distance",   required_argument, 0,   0},
  {"help",       no_argument,       0, 'h'},
  {"missing",    required_argument, 0,   0},
  {"nogap",      no_argument,       0,   0},
  {"noconstant", no_argument,       0,   0},
  {"ordering",   required_argument, 0,   0},  
  {"reorder",    no_argument,       0, 'r'},
  {"silent",     no_argument,       0, 's'},  
  {"shuffles",   required_argument, 0,   0},
  {"smooth",     required_argument, 0,   0},
  {"seqtype",    required_argument, 0,   0},
  {"verbose",    no_argument,       0, 'v'},
  {NULL, 0, NULL, 0}
};

/* forward declaration(s) of private function(s) */
static void initializeOptions (char *name);
static void processOptions(int argc, char *argv[]);
static void usage(int status);
static void warn ( char *fmt, ... );
static void serve_Viennese_melange(void);

/* ==> public function(s) <== */
/*
 */
OptionS *
decodeCML(int argc, char *argv[]) {
  initializeOptions(argv[0]);
  processOptions(argc, argv);
  return (&OptS);
}

char*
optionString(void)
{
  static char options[200];
  *options = '\0';

  sprintf(options+strlen(options),   "--cutoff=%.2f ", OptS.CutOff);
  sprintf(options+strlen(options),  "--distance=%s ",  OptS.DistMethod);  
  if ( OptS.DistMx[0] != '\0' )
    sprintf(options+strlen(options),   "--matrix=%s ", OptS.DistMx);    
  sprintf(options+strlen(options),    "--missing=%s ", OptS.Missing);
  sprintf(options+strlen(options),      "--nogap=%d ", OptS.NoGap);
  sprintf(options+strlen(options), "--noconstant=%d ", OptS.NoConstant);
  sprintf(options+strlen(options),   "--ordering=%s ", OptS.OrdMethod);
  sprintf(options+strlen(options),   "--shuffles=%d ", OptS.Shuffles);
  sprintf(options+strlen(options),     "--smooth=%d ", OptS.Smooth);
  sprintf(options+strlen(options),    "--seqtype=%c ", OptS.SeqType); 

  return (options);
}

/* ==> private function(s) <== */

/*
  print usage information to stderr and exit
*/
static void
usage(int status)
{
  char dmx[256] = "calculate distances with nnet";
  if ( OptS.DistMx[0] != '\0' ) strcpy(dmx, OptS.DistMx);

  fprintf(stderr, "\nUsage: %s [OPTIONS] MultiFastaAlignmen\n", OptS.ExeName);
  fprintf(stderr,
          "Options:\n"
	  " --cutoff FLOAT     Set cutoff value to FLOAT.\n"
	  "                     (default is >%.2f<)\n"
          " --distance STRING  Distance methode used by NeighbotNet.\n"
          "                     (default is >%s< [HAMMING|GTR])\n"	  
	  " -h, --help         Print usage message for %s.\n"
          " --matrix FILE      Use distance matrix for NeighbotNet from FILE.\n"
          "                     (default is >%s<)\n"
	  " --missing STRING   Set list of missing characters to STRING.\n"
	  "                     (default is >%s<)\n",
	  OptS.CutOff, OptS.DistMethod, OptS.ExeName, dmx, OptS.Missing);
  fprintf(stderr,
	  " --nogap            Don't count gap symbol as character state.\n"
	  "                     (default is >%d<)\n"
	  " --noconstant       Ignore constant columns in output alignment.\n"
          "                     (default is >%d<)\n"
          " --ordering STRING  Choose ordering method.\n"
          "                     (default is >%s<)\n"
	  "                      nnet         NeighborNet\n"
	  "                      qnet         QNet\n",
	  OptS.NoGap, OptS.NoConstant, OptS.OrdMethod);

  fprintf(stderr,
	  "                      rand[,INT]   sample INT random permutation\n"
          "                                   (default is >%d<)\n"
	  "                      INT,INT,...  provide ordering explicitly as\n"
          "                                   comma-seperated string of INTs\n"
	  "                      all          all permutations\n"
          " -r, --reorder      Only reorder alignment rows (do not shuffle)\n"
          "                     (default is >%d<)\n"
          " --shuffles INT     Perform INT random shuffles per column.\n"
          "                     (default is >%d<)\n",
	  OptS.SampleSize, OptS.Shuffles, OptS.ReOrder);
  
  fprintf(stderr,
	  " -s, --silent       Turn off status bar\n"
          "                     (default is >%d<)\n"
          " --smooth INT       Running avarage over INT columns.\n"
          "                     (default is >%d<)\n"
          " --seqtype CHAR     Set sequence type of input alignment to CHAR\n"
          "                     (default is >%c<) [D=DNA, P=Protein, R=RNA]\n"
	  " -v, --verbose      Verbose mode. Causes %s to print debugging\n"
	  "                    messages about its progress.\n\n",
	  OptS.Silent, OptS.Smooth, OptS.SeqType, OptS.ExeName);

  exit (status);
}

/*
  call example:
  warn("calloc(): out of memory when requesting %d bytes", numbytes)
 */
static void
warn ( char *fmt, ... )
{
  va_list args;
  
  va_start( args, fmt );
  fprintf( stderr, "\nWARNING: ");
  vfprintf( stderr, fmt, args);
  fprintf( stderr,"\n");
  va_end( args );
}

/**
 * Set options to default value(s)
 */
static void
initializeOptions(char *exename)
{
  *(OptS.ExeName) = '\0';
  *(OptS.InName)  = '\0';
  *(OptS.Missing) = '\0';
  *(OptS.DistMx)  = '\0';
  strcpy(OptS.ExeName, exename);
  OptS.SeqType = 'D';
  strcpy(OptS.DistMethod, "HAMMING");
  strcpy(OptS.OrdMethod, "nnet");
  OptS.CutOff     = 0.8;
  strcpy(OptS.Missing, "N");
  OptS.NoGap      = 0;
  OptS.NoConstant = 0;
  OptS.Silent     = 0;
  OptS.Shuffles   = 1000;
  OptS.ReOrder    = 0;
  OptS.Smooth     = 1;
  OptS.Verbose    = 0;
  OptS.SampleSize = 1000;
}

/**
 * Process command-line options and update OptS structure
 */
static void
processOptions(int argc, char *argv[]) {
  int c, option_index = 0;

  while ((c = getopt_long (argc, argv, "hrsv", 
			   long_options, &option_index)) != EOF) {
    switch (c) {
    case 0:

      if ( strcmp(long_options[option_index].name, "Kaffee") == 0 ) {
	serve_Viennese_melange();
	exit(EXIT_SUCCESS);
      }
      
      if ( strcmp(long_options[option_index].name, "cutoff") == 0 ) {
	double tmp;
	if ( sscanf(optarg, "%lf", &tmp ) == 0 ) {
	  warn ("Option requires argument: No cutoff value specified");
	  usage (EXIT_FAILURE);
	}
	else { OptS.CutOff = tmp; }
      }

      if ( strcmp(long_options[option_index].name, "distance" ) == 0 ) {
	char tmp[256];
	if ( sscanf(optarg, "%s", tmp) == 0 ) {
	  warn ("Option requires argument: No distance method specified");
	  usage (EXIT_FAILURE);
	}
	else {
	  if (strcmp(tmp, "HAMMING")==0 || strcmp(tmp, "GTR")==0) {
	    strcpy(OptS.DistMethod, tmp);	    
	  }
	  else {
	    warn ("Unknown distance method: `%s'", tmp);
	    usage (EXIT_FAILURE);
	  }
	}
      }

      if ( strcmp(long_options[option_index].name, "matrix" ) == 0 ) {
	char tmp[256];
	if ( sscanf(optarg, "%s", tmp) == 0 ) {
	  warn ("Option requires argument: No distance matrix file specified");
	  usage (EXIT_FAILURE);
	}
	else { strcpy(OptS.DistMx, tmp); }
      }

      if ( strcmp(long_options[option_index].name, "missing" ) == 0 ) {
	char tmp[256];
	if ( sscanf(optarg, "%s", tmp) == 0 ) {
	  warn ("Option requires argument: No missing characters specified");
	  usage (EXIT_FAILURE);
	}
	else {
	  char *c = NULL;

	  /* erase '-' from missing symbols string */
	  while ((c = strchr(tmp, '-')) != NULL)
	    strcpy(c,c+1);
	  
	  strcpy(OptS.Missing, tmp);
	}
      }

      if ( strcmp(long_options[option_index].name, "nogap") == 0 ) {
	OptS.NoGap = 1;
      }

      if ( strcmp(long_options[option_index].name, "noconstant") == 0 ) {
	/* OptS.NoConstant = 1; only for testing */
	OptS.NoConstant++;
      }

     if ( strcmp(long_options[option_index].name, "ordering" ) == 0 ) {
	char tmp[256];
	if ( sscanf(optarg, "%s", tmp) == 0 ) {
	  warn ("Option requires argument: No ordering method specified");
	  usage (EXIT_FAILURE);
	}
	else {
	  /*
	  if ((strcmp(tmp, "nnet") == 0) || (strcmp(tmp, "qnet") == 0)) {
	    warn ("Unknown ordering method: `%s'", tmp);
	    usage (EXIT_FAILURE);
	  }
	  */
	  strcpy(OptS.OrdMethod, tmp);
	}
     }
      
     if ( strcmp(long_options[option_index].name, "reorder") == 0 ) {
       OptS.ReOrder = 1;
     }

     if ( strcmp(long_options[option_index].name, "shuffles" ) == 0 ) {
       int tmp;
       if ( sscanf(optarg, "%d", &tmp) == 0 ) {
	 warn ("Option requires argument: No number of shufflings specified");
	 usage (EXIT_FAILURE);
       }
       else { OptS.Shuffles = tmp; }
     }
     
     if ( strcmp(long_options[option_index].name, "silent") == 0 ) {
       OptS.Silent = 1;
     }
      
     if ( strcmp(long_options[option_index].name, "smooth" ) == 0 ) {
       int tmp;
       if ( sscanf(optarg, "%d", &tmp) == 0 ) {
	 warn ("Option requires argument: No smoothing value specified");
	 usage (EXIT_FAILURE);
       }
       else { OptS.Smooth = tmp; }
     }

     if ( strcmp(long_options[option_index].name, "seqtype" ) == 0 ) {
       char tmp;
       if ( sscanf(optarg, "%c", &tmp) == 0 ) {
	 warn ("Option requires argument: No sequence type specified");
	 usage (EXIT_FAILURE);
       }
       else { OptS.SeqType = tmp; }
     }
      
     if ( strcmp(long_options[option_index].name, "verbose") == 0 ) {
       OptS.Verbose++;
     }
      
     break;
    case 'h':
      if (OptS.NoGap == 1) strcat(OptS.Missing, "-");
      usage (EXIT_SUCCESS);
      break;
    case 'r':
      OptS.ReOrder = 1;
      break;
    case 's':
      OptS.Silent = 1;
      break;
    case 'v':
      OptS.Verbose++;
      break;
    default:
      usage (EXIT_FAILURE);
    }
  }

  /* add `-' to list of missing symbols if option --nogap was given */
  if (OptS.NoGap == 1) strcat(OptS.Missing, "-");

  /* parse the multifasta input file name from the command line */
  if (argv[optind] == NULL) {
    warn ("No input multiple alignment file specified.");
    exit (EXIT_FAILURE);
  }
  else {
    strcpy(OptS.InName,  argv[optind]);
  }

#if 0
  /* parse the to file names from the remaining command-line */
  if ( (optind + 2) <= argc) {
    strcpy(OptS.PhylipFile,  argv[optind++]);
    strcpy(OptS.DumpFile,    argv[optind++]);
  }
  else {
    warn ("Not enough Input Files. (Need a Phylip AND a Dump File)");
    usage(EXIT_FAILURE);
  }
#endif
}

static void
serve_Viennese_melange(void)
{
  int i;
  char melange[16][19] = {
    "                  ",
    "   Bitt scheeen   ",
    "                  ",
    "    )) )) ))      ",
    "   (( (( ((       ",
    "    )) )) ))      ",
    " #99999999999#### ",
    " #           #  # ",
    " #           ###  ",
    "  #         #     ",
    "   #       #      ",
    "    ######        ",
    "                  ",
    " a wiina meloosch ",
    "    sche haas!    ",
    "                  "};
  for ( i=0; i<16; i++)
    fprintf(stderr,"%s\n", melange[i]);
  fflush(stderr);
}

/* End of file */
