#!/bin/sh
# -*-Bash-*-
# Last changed Time-stamp: <2006-05-30 16:29:46 xtof>
# $Id: roboti.sh,v 1.1 2006/05/30 14:36:07 xtof Exp $
#

# strip leading path and tailing extension(s)
function to_basename {
    local basename=${1%/}
    basename=${basename##*/}
    basename=${basename%%.*}
    echo $basename
}

# set global(s) variables to default value(s)
NPATH=.
DIR=.
EXT=fas
INC=0.1
USAGE="Usage: $0 [-d Str] [-e Str] [-i Flt] [-n Str]"

# process commandline options
while getopts "d:e:i:n:h" opt; do
    case $opt in
	d)
	    DIR=$OPTARG
	    ;;
	e)
	    EXT=$OPTARG
	    ;;
	h)
	    echo""
            echo $USAGE
	    echo "       -d Str  path to sequence files"
	    echo "               (default: $DIR)"
	    echo "       -e Str  extension of sequence files"
	    echo "               (default: $EXT)"
	    echo "       -i Flt  set increment for cutoff"
	    echo "               (default: $INC)"
	    echo "       -n Flt  path to program noisy"
	    echo "               (default: $NPATH)"
	    echo "       -h      display this help message"
            echo ""
	    exit 1
	    ;;
	i)
	    INC=$OPTARG
	    ;;
	n)
	    NPATH=$OPTARG
	    ;;
	\?)
	echo $USAGE
	exit 1
	;;
    esac
done
shift $(($OPTIND - 1))

# check if program noisy is there
if [ ! -x $NPATH/noisy ]; then
 echo ""
 echo "ERROR: Can't find program noisy in path $NPATH"
 echo "       You can turn the correct path over to me with the -n option"
 echo ""
 exit 1
fi

#
# main loop
#

# iterate over all files
for file in $(ls $DIR/*.$EXT)
do
 basename=$(to_basename $file)
 echo "processing $file ..."
 # forach file run noisy with various cutoff values 
 for cutoff in $(seq 0 $INC 1)
 do
  echo "cutoff = $cutoff"
  $NPATH/noisy --silent --cutoff=$cutoff $file
  mv ${basename}_sta.gr ${basename}_stat${cutoff}.gr
  mv ${basename}_out.fas ${basename}_out${cutoff}.fas
 done
done

# End of file
